@@ -2,25 +2,26 @@
 /* vim:set et sts=4: */
 /* IBus - The Input Bus
  * Copyright (C) 2008-2010 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2008-2010 Red Hat, Inc.
+ * Copyright (C) 2018 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2008-2018 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
+ * USA
  */
-#include <dbus/dbus.h>
 #include "ibushotkey.h"
+#include "ibusmarshalers.h"
 #include "ibuskeysyms.h"
 #include "ibusinternal.h"
 #include "ibusshare.h"
@@ -60,25 +61,24 @@ static IBusHotkey   *ibus_hotkey_new                (guint                   key
                                                      guint                   modifiers);
 static IBusHotkey   *ibus_hotkey_copy               (const IBusHotkey       *src);
 static void          ibus_hotkey_free               (IBusHotkey             *hotkey);
-/*
-static gboolean      ibus_hotkey_serialize          (IBusHotkey             *hotkey,
-                                                     IBusMessageIter        *iter);
-static gboolean      ibus_hotkey_deserialize        (IBusHotkey             *hotkey,
-                                                     IBusMessageIter        *iter);
-*/
-static void          ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *klass);
+static void          ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *class);
 static void          ibus_hotkey_profile_init       (IBusHotkeyProfile      *profile);
 static void          ibus_hotkey_profile_destroy    (IBusHotkeyProfile      *profile);
 static gboolean      ibus_hotkey_profile_serialize  (IBusHotkeyProfile      *profile,
-                                                     IBusMessageIter        *iter);
-static gboolean      ibus_hotkey_profile_deserialize(IBusHotkeyProfile      *profile,
-                                                     IBusMessageIter        *iter);
+                                                     GVariantBuilder        *builder);
+static gint          ibus_hotkey_profile_deserialize(IBusHotkeyProfile      *profile,
+                                                     GVariant               *variant);
 static gboolean      ibus_hotkey_profile_copy       (IBusHotkeyProfile      *dest,
                                                      const IBusHotkeyProfile*src);
 static void          ibus_hotkey_profile_trigger    (IBusHotkeyProfile      *profile,
                                                      GQuark                  event,
                                                      gpointer                user_data);
 
+// Normalize modifiers by setting necessary modifier bits according to keyval.
+static guint         normalize_modifiers            (guint                   keyval,
+                                                     guint                   modifiers);
+static gboolean      is_modifier                    (guint                   keyval);
+
 static IBusSerializableClass *parent_class = NULL;
 
 static guint profile_signals[LAST_SIGNAL] = { 0 };
@@ -97,40 +97,6 @@ ibus_hotkey_get_type (void)
     return type;
 }
 
-/*
-static gboolean
-ibus_hotkey_serialize (IBusHotkey      *hotkey,
-                       IBusMessageIter *iter)
-{
-    gboolean retval;
-
-    retval = ibus_message_iter_append (iter, G_TYPE_UINT, &hotkey->keyval);
-    g_return_val_if_fail (retval, FALSE);
-
-    retval = ibus_message_iter_append (iter, G_TYPE_UINT, &hotkey->modifiers);
-    g_return_val_if_fail (retval, FALSE);
-
-    return TRUE;
-}
-
-static gboolean
-ibus_hotkey_deserialize (IBusHotkey      *hotkey,
-                         IBusMessageIter *iter)
-{
-    gboolean retval;
-
-    retval = ibus_message_iter_get (iter, G_TYPE_UINT, &hotkey->keyval);
-    g_return_val_if_fail (retval, FALSE);
-    ibus_message_iter_next (iter);
-
-    retval = ibus_message_iter_get (iter, G_TYPE_UINT, &hotkey->modifiers);
-    g_return_val_if_fail (retval, FALSE);
-    ibus_message_iter_next (iter);
-
-    return TRUE;
-}
-*/
-
 static IBusHotkey *
 ibus_hotkey_new (guint keyval,
                  guint modifiers)
@@ -185,7 +151,7 @@ ibus_hotkey_profile_get_type (void)
         (GBaseInitFunc)     NULL,
         (GBaseFinalizeFunc) NULL,
         (GClassInitFunc)    ibus_hotkey_profile_class_init,
-        NULL,               /* class finialize */
+        NULL,               /* class finalize */
         NULL,               /* class data */
         sizeof (IBusHotkeyProfile),
         0,
@@ -203,14 +169,14 @@ ibus_hotkey_profile_get_type (void)
 }
 
 static void
-ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *klass)
+ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *class)
 {
-    IBusObjectClass *object_class = IBUS_OBJECT_CLASS (klass);
-    IBusSerializableClass *serializable_class = IBUS_SERIALIZABLE_CLASS (klass);
+    IBusObjectClass *object_class = IBUS_OBJECT_CLASS (class);
+    IBusSerializableClass *serializable_class = IBUS_SERIALIZABLE_CLASS (class);
 
-    parent_class = (IBusSerializableClass *) g_type_class_peek_parent (klass);
+    parent_class = (IBusSerializableClass *) g_type_class_peek_parent (class);
 
-    g_type_class_add_private (klass, sizeof (IBusHotkeyProfilePrivate));
+    g_type_class_add_private (class, sizeof (IBusHotkeyProfilePrivate));
 
     object_class->destroy = (IBusObjectDestroyFunc) ibus_hotkey_profile_destroy;
 
@@ -218,9 +184,7 @@ ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *klass)
     serializable_class->deserialize = (IBusSerializableDeserializeFunc) ibus_hotkey_profile_deserialize;
     serializable_class->copy        = (IBusSerializableCopyFunc) ibus_hotkey_profile_copy;
 
-    klass->trigger = ibus_hotkey_profile_trigger;
-
-    g_string_append (serializable_class->signature, "av");
+    class->trigger = ibus_hotkey_profile_trigger;
 
     /* install signals */
     /**
@@ -236,11 +200,11 @@ ibus_hotkey_profile_class_init (IBusHotkeyProfileClass *klass)
      */
     profile_signals[TRIGGER] =
         g_signal_new (I_("trigger"),
-            G_TYPE_FROM_CLASS (klass),
+            G_TYPE_FROM_CLASS (class),
             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
             G_STRUCT_OFFSET (IBusHotkeyProfileClass, trigger),
             NULL, NULL,
-            ibus_marshal_VOID__UINT_POINTER,
+            _ibus_marshal_VOID__UINT_POINTER,
             G_TYPE_NONE,
             2,
             G_TYPE_UINT,
@@ -262,6 +226,8 @@ ibus_hotkey_profile_init (IBusHotkeyProfile *profile)
     priv->mask = IBUS_SHIFT_MASK |
                  IBUS_CONTROL_MASK |
                  IBUS_MOD1_MASK |
+                 IBUS_SUPER_MASK |
+                 IBUS_HYPER_MASK |
                  IBUS_RELEASE_MASK;
 }
 
@@ -294,26 +260,26 @@ ibus_hotkey_profile_destroy (IBusHotkeyProfile *profile)
 
 static gboolean
 ibus_hotkey_profile_serialize (IBusHotkeyProfile *profile,
-                               IBusMessageIter   *iter)
+                               GVariantBuilder   *builder)
 {
     gboolean retval;
 
-    retval = parent_class->serialize ((IBusSerializable *) profile, iter);
+    retval = parent_class->serialize ((IBusSerializable *) profile, builder);
     g_return_val_if_fail (retval, FALSE);
 
     return TRUE;
 }
 
-static gboolean
+static gint
 ibus_hotkey_profile_deserialize (IBusHotkeyProfile *profile,
-                                 IBusMessageIter   *iter)
+                                 GVariant          *variant)
 {
-    gboolean retval;
+    gint retval;
 
-    retval = parent_class->deserialize ((IBusSerializable *) profile, iter);
-    g_return_val_if_fail (retval, FALSE);
+    retval = parent_class->deserialize ((IBusSerializable *) profile, variant);
+    g_return_val_if_fail (retval, 0);
 
-    return TRUE;
+    return retval;
 }
 
 static gboolean
@@ -348,6 +314,56 @@ ibus_hotkey_profile_trigger (IBusHotkeyProfile *profile,
     // g_debug ("%s is triggerred", g_quark_to_string (event));
 }
 
+static guint
+normalize_modifiers (guint keyval,
+                     guint modifiers)
+{
+    switch(keyval) {
+    case IBUS_KEY_Control_L:
+    case IBUS_KEY_Control_R:
+        return modifiers | IBUS_CONTROL_MASK;
+    case IBUS_KEY_Shift_L:
+    case IBUS_KEY_Shift_R:
+        return modifiers | IBUS_SHIFT_MASK;
+    case IBUS_KEY_Alt_L:
+    case IBUS_KEY_Alt_R:
+    // Chrome OS does not have Meta key. Instead, shift+alt generates Meta keyval.
+    case IBUS_KEY_Meta_L:
+    case IBUS_KEY_Meta_R:
+        return modifiers | IBUS_MOD1_MASK;
+    case IBUS_KEY_Super_L:
+    case IBUS_KEY_Super_R:
+        return modifiers | IBUS_SUPER_MASK;
+    case IBUS_KEY_Hyper_L:
+    case IBUS_KEY_Hyper_R:
+        return modifiers | IBUS_HYPER_MASK;
+    default:
+        return modifiers;
+    }
+}
+
+static gboolean
+is_modifier (guint keyval)
+{
+    switch(keyval) {
+    case IBUS_KEY_Control_L:
+    case IBUS_KEY_Control_R:
+    case IBUS_KEY_Shift_L:
+    case IBUS_KEY_Shift_R:
+    case IBUS_KEY_Alt_L:
+    case IBUS_KEY_Alt_R:
+    case IBUS_KEY_Meta_L:
+    case IBUS_KEY_Meta_R:
+    case IBUS_KEY_Super_L:
+    case IBUS_KEY_Super_R:
+    case IBUS_KEY_Hyper_L:
+    case IBUS_KEY_Hyper_R:
+        return TRUE;
+    default:
+        return FALSE;
+    }
+}
+
 gboolean
 ibus_hotkey_profile_add_hotkey (IBusHotkeyProfile *profile,
                                 guint              keyval,
@@ -357,7 +373,8 @@ ibus_hotkey_profile_add_hotkey (IBusHotkeyProfile *profile,
     IBusHotkeyProfilePrivate *priv;
     priv = IBUS_HOTKEY_PROFILE_GET_PRIVATE (profile);
 
-    IBusHotkey *hotkey = ibus_hotkey_new (keyval, modifiers);
+    IBusHotkey *hotkey = ibus_hotkey_new (keyval,
+                                          normalize_modifiers (keyval, modifiers & priv->mask));
 
     /* has the same hotkey in profile */
     if (g_tree_lookup (priv->hotkeys, hotkey) != NULL) {
@@ -410,6 +427,8 @@ ibus_hotkey_profile_remove_hotkey (IBusHotkeyProfile *profile,
     IBusHotkeyProfilePrivate *priv;
     priv = IBUS_HOTKEY_PROFILE_GET_PRIVATE (profile);
 
+    modifiers = normalize_modifiers (keyval, modifiers & priv->mask);
+
     IBusHotkey hotkey = {
         .keyval = keyval,
         .modifiers = modifiers
@@ -435,7 +454,7 @@ ibus_hotkey_profile_remove_hotkey (IBusHotkeyProfile *profile,
             break;
     }
 
-    g_assert (p2->event == event);
+    g_assert (p2 && p2->event == event);
 
     p2->hotkeys = g_list_remove (p2->hotkeys, p1);
     if (p2->hotkeys == NULL) {
@@ -487,13 +506,35 @@ ibus_hotkey_profile_filter_key_event (IBusHotkeyProfile *profile,
     IBusHotkeyProfilePrivate *priv;
     priv = IBUS_HOTKEY_PROFILE_GET_PRIVATE (profile);
 
+    modifiers = normalize_modifiers (keyval, modifiers & priv->mask);
+    prev_modifiers = normalize_modifiers (prev_keyval, prev_modifiers & priv->mask);
+
     IBusHotkey hotkey = {
         .keyval = keyval,
-        .modifiers = modifiers & priv->mask,
+        .modifiers = modifiers,
     };
 
-    if ((modifiers & IBUS_RELEASE_MASK) && keyval != prev_keyval) {
-        return 0;
+    if (modifiers & IBUS_RELEASE_MASK) {
+        /* previous key event must be a press key event */
+        if (prev_modifiers & IBUS_RELEASE_MASK)
+            return 0;
+
+        /* modifiers should be same except the release bit */
+        if (modifiers != (prev_modifiers | IBUS_RELEASE_MASK))
+            return 0;
+
+        /* If it is release key event,
+         * we need check if keyval is equal to the prev keyval.
+         * If keyval is not equal to the prev keyval,
+         * but both keyvals are modifier keys,
+         * we will still search it in hotkeys.
+         * It is for matching some key sequences like:
+         * Shift Down, Alt Down, Shift Up => Shift+Alt+Release - Shift hotkey
+         **/
+        if ((keyval != prev_keyval) &&
+            (is_modifier (keyval) == FALSE ||
+             is_modifier (prev_keyval) == FALSE))
+            return 0;
     }
 
     GQuark event = (GQuark) GPOINTER_TO_UINT (g_tree_lookup (priv->hotkeys, &hotkey));
@@ -513,9 +554,11 @@ ibus_hotkey_profile_lookup_hotkey (IBusHotkeyProfile *profile,
     IBusHotkeyProfilePrivate *priv;
     priv = IBUS_HOTKEY_PROFILE_GET_PRIVATE (profile);
 
+    modifiers = normalize_modifiers (keyval, modifiers & priv->mask);
+
     IBusHotkey hotkey = {
         .keyval = keyval,
-        .modifiers = modifiers & priv->mask,
+        .modifiers = modifiers,
     };
 
     return (GQuark) GPOINTER_TO_UINT (g_tree_lookup (priv->hotkeys, &hotkey));
